/**********************************************************************
 * This file is part of FMR (Fast Machine Reassignment)               
 *                                                                   
 * ROADEF/EURO Challenge 2012: Machine Reassignment                 
 * http://challenge.roadef.org/2012/en/				   
 * Team S26
 *     
 * Copyright (C) 2011-12 Laboratoire d'Informatique de Paris Nord 
 * UMR CNRS 7030 
 * Institut Galilée - Université Paris-Nord
 * 99, avenue Jean-Baptiste Clément 
 * 93430 Villetaneuse  
 *        
 * FMR is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *  
 * FMR is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with FMR; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **********************************************************************/
#include "fmr.h"
#include "fmr_ens.h"
#include "fmr_tools.h"
#include "fmr_composanteConnexe.h"

/* Identification des composantes connexes : compsantes[i] donne l'ensemble des services de la ieme comp. */
void
compConnexe (unsigned int *maxcomp, unsigned int *maxcomp2) {
	/* a gerer au niveau global ou a libérer (free) à la fin de la fonction */
	Ensemble_t composantes[MAXNBSERVICES];
   	unsigned int nbComposantes=0;
	unsigned int i,serviceCourant=0,s,stempo;
	Ensemble_t 
		travail = Ens_nouveau(instance.nb_services),
		restant_afaire= Ens_nouveau(instance.nb_services);

	/* initialisation */
	Ens_rempli(restant_afaire,instance.nb_services);
	Ens_zero(travail, instance.nb_services);

	do {
		composantes[nbComposantes] = Ens_nouveau(instance.nb_services);
		Ens_zero(composantes[nbComposantes], instance.nb_services);

		Ens_ote(serviceCourant, restant_afaire);
		Ens_ajoute(serviceCourant, composantes[nbComposantes]);
		stempo=serviceCourant;
		do {
			/*  predecesseurs de stempo sont dans la composante */
			for (i=0; i< instance.services[stempo].nb_dep;i++) {
				s= instance.services[stempo].dependances[i];
				Ens_ajoute(s, composantes[nbComposantes]);
				if (Ens_appartient(s, restant_afaire)) {
					Ens_ajoute(s, travail);
					Ens_ote(s, restant_afaire);
				}
			}

			/* successeurs de stempo sont dans la composante*/
			for (i=0; i< instance.services[stempo].nb_succ; i++) {
				s=instance.services[stempo].successeurs[i];
				Ens_ajoute(s, composantes[nbComposantes]);
				if (Ens_appartient(s, restant_afaire)) {
					Ens_ajoute(s, travail);
					Ens_ote(s, restant_afaire);
				}
			}
		} while ( (stempo=Ens_ote_prem_elt(travail, instance.nb_services)) != (unsigned int)-1); /* tant que non vide */
		
		nbComposantes++;
	} while ( (serviceCourant=Ens_elt_suivant(serviceCourant,restant_afaire, instance.nb_services)) > 0); /* idem */

	*maxcomp=0;
	*maxcomp2=0;
	for (i=0; i< nbComposantes; i++) {
		stempo= Ens_card(composantes[i], instance.nb_services);  
		if (*maxcomp < stempo) {
			*maxcomp2= *maxcomp;
			*maxcomp= stempo;
		} else if (*maxcomp2 < stempo) {
			*maxcomp2 = stempo;
		}
	}
	DEBUG_STATS_PRINT("Nbre composantes connexes:%d, maxcomp=%u, maxcomp2=%u\n", nbComposantes,*maxcomp,*maxcomp2);
//#ifdef DEBUGIO
	for (i=0; i< nbComposantes; i++) {
		if (Ens_card(composantes[i], instance.nb_services) >1) {
		  DEBUG_STATS_PRINT("composante %d: card=%d \n ", i, Ens_card(composantes[i], instance.nb_services));
		}
		//Ens_aff(composantes[i], instance.nb_services);
		//DEBUG_STATS_PRINT("\n");
	}
//#endif	
	/* On suppose que composantes[] n'est pas à utiliser ailleurs ? */
	for( i=0; i< nbComposantes; i++) {
		free(composantes[i]);
	}
	free (travail);
	free (restant_afaire);
}

