/**********************************************************************
 * This file is part of FMR (Fast Machine Reassignment)               
 *                                                                   
 * ROADEF/EURO Challenge 2012: Machine Reassignment                 
 * http://challenge.roadef.org/2012/en/				   
 * Team S26
 *     
 * Copyright (C) 2011-12 Laboratoire d'Informatique de Paris Nord 
 * UMR CNRS 7030 
 * Institut Galilée - Université Paris-Nord
 * 99, avenue Jean-Baptiste Clément 
 * 93430 Villetaneuse  
 *        
 * FMR is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *  
 * FMR is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with FMR; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include "fmr_heap.h"

	/* Arbre binaire quasi complet pour stocker les meilleurs mouvements */
void ajoutHeap(unsigned short u1, unsigned short u2, long gain, deplt_t** t, unsigned *nbElt) {
	deplt_t *nouv ;
	unsigned int courant = *nbElt;
	unsigned int pere;
        if (*nbElt>=MAXNBDEPL) {
		if (t[courant-1]->gainLoad>= gain)
                   return;
		else { // meilleur que le plus mauvais : on écrase le plus mauvais
		   courant= --(*nbElt);
	           nouv=t[courant];
		}
	} else  {
		nouv= (deplt_t *)malloc(sizeof(deplt_t));
	}

	nouv->u1= u1;
	nouv->u2= u2;
	nouv->gainLoad=gain;

	pere=(courant-1)/2;
	while (courant>=1 && t[pere]->gainLoad < gain) {
		t[courant]=t[pere];
		courant=pere;
		pere=(courant-1)/2;
	}
  	t[courant]=nouv;
	(*nbElt)++;
}

void ajoutHeap2(unsigned short u1, unsigned short u2, long gain, unsigned int indice[], deplt_t t[], unsigned *nbElt) {
	unsigned int nouv;
	unsigned int courant= *nbElt;
	unsigned int pere;
	if (*nbElt>=MAXNBDEPL) {
                if (t[indice[courant-1]].gainLoad>= gain)
			return;
		else {
			courant= --(*nbElt);
			nouv=indice[courant];
		}
	} else {
		nouv=courant;
	}
	t[nouv].u1=u1;
	t[nouv].u2=u2;
	t[nouv].gainLoad=gain;

	pere=(courant-1)/2;
        while (courant>=1 && t[indice[pere]].gainLoad < gain) {
		indice[courant]=indice[pere];
		courant=pere;
		pere=(courant-1)/2;
	}
	indice[courant]=nouv;
	(*nbElt)++;
}

/*retire l'element racine (le meilleur donc) du heap*/
void oteHeap (deplt_t** t, unsigned *nbElt) {
	unsigned int courant = 0;
	deplt_t *ptr;
	long meilleurFils;
	unsigned int posmeilleurFils,j;

	if (t[0]==NULL)
		fprintf(stderr,"erreur oteHeap sur NULL, nbElt=%u\n", *nbElt);
	else
		free(t[0]);
	*nbElt -= 1;
	t[0]= t[*nbElt];
	
	posmeilleurFils = courant;
	meilleurFils = t[courant]->gainLoad;
	j=2*courant+1;

	if (j < *nbElt && t[j]->gainLoad > meilleurFils) {
		posmeilleurFils= j;
	}
	j++;
	if (j < *nbElt && t[j]->gainLoad > meilleurFils) {
		posmeilleurFils= j;
	}
	meilleurFils = t[posmeilleurFils]->gainLoad;

	while (t[courant]->gainLoad < meilleurFils) { /* si on a amélioré c'est qu'il y a forcément 1 fils */
	 	ptr = t[courant];
		t[courant] = t[posmeilleurFils];
		t[posmeilleurFils] = ptr;	
		courant=posmeilleurFils;
		meilleurFils = t[courant]->gainLoad;
		j=2*courant+1;
	        if (j < *nbElt && t[j]->gainLoad > meilleurFils) {
			posmeilleurFils= j;
		}
		j++;
        	if (j < *nbElt && t[j]->gainLoad > meilleurFils) {
			posmeilleurFils= j;
		}
        	meilleurFils = t[posmeilleurFils]->gainLoad;
	}
}

void oteHeap2 (unsigned indice[], deplt_t t[], unsigned *nbElt) {
	unsigned int courant = 0,j;
	long gainMeilleurFils;
	unsigned int posmeilleurFils,k;

	if (nbElt == 0) return;
	--(*nbElt);
	indice[0] = indice[*nbElt]; // on écrase le meilleur avec le pire
	posmeilleurFils=0;
	gainMeilleurFils=t[indice[0]].gainLoad;
	j=2*courant+1;
	if (j < *nbElt && t[indice[j]].gainLoad > gainMeilleurFils)
		posmeilleurFils = j;
	j++;
	if (j < *nbElt && t[indice[j]].gainLoad > gainMeilleurFils)
                posmeilleurFils = j;
	gainMeilleurFils= t[indice[posmeilleurFils]].gainLoad;

	while (t[indice[courant]].gainLoad  < gainMeilleurFils) {
		k=indice[courant]; indice[courant]=indice[posmeilleurFils]; 
		indice[posmeilleurFils]=k;
		gainMeilleurFils=t[indice[courant]].gainLoad;
		j=2*courant+1;
		if (j < *nbElt && t[indice[j]].gainLoad > gainMeilleurFils)
			posmeilleurFils = j;
		j++;
		if (j < *nbElt && t[indice[j]].gainLoad > gainMeilleurFils)
                        posmeilleurFils = j;
		gainMeilleurFils= t[indice[posmeilleurFils]].gainLoad;
	}
}

/* bourrin : libere les struct qui parlent du ps p */
void oteprocHeap (unsigned int p, deplt_t **t, unsigned nbELt) {
	unsigned int i;

	for (i=0; i< nbELt; i++) {
		if (t[i]!=NULL && t[i]->u1 == p) {
			free(t[i]);
			t[i]=NULL;
		}
	}
}

void free_heap_elts(deplt_t **t, unsigned nbELt) {
	unsigned int i;
	for (i=0; i<nbELt; i++) {
		if (t[i]!=NULL)
			free(t[i]);
	}
}

void afficheHeap( deplt_t** t, unsigned nbElt) {
   unsigned i;

	for (i=0;i<nbElt;i++) 
		printf("G=%ld u1=%hu u2=%hu\n",t[i]->gainLoad, t[i]->u1, t[i]->u2);
}
