/**********************************************************************
 * This file is part of FMR (Fast Machine Reassignment)               
 *                                                                   
 * ROADEF/EURO Challenge 2012: Machine Reassignment                 
 * http://challenge.roadef.org/2012/en/				   
 * Team S26
 *     
 * Copyright (C) 2011-12 Laboratoire d'Informatique de Paris Nord 
 * UMR CNRS 7030 
 * Institut Galilée - Université Paris-Nord
 * 99, avenue Jean-Baptiste Clément 
 * 93430 Villetaneuse  
 *        
 * FMR is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *  
 * FMR is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with FMR; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **********************************************************************/

#include <string.h>

#include "fmr.h"
#include "fmr_io.h"
#include "fmr_tools.h"

#ifdef DEBUGIO
#include "fmr_constraints.h"
#endif

void modif_sc(){
    unsigned int i, j;
    unsigned int seuil=0;
    unsigned int diffCSC[instance.nb_machines];
    
    for( i = 0 ; i < instance.nb_machines ; i++ ) {
        diffCSC[i]=0;
        for( j = 0 ; j < instance.nb_ressources ; j++ ) {
            diffCSC[i]+=instance.machines[i].C[j]-instance.machines[i].SC[j];
        }
        seuil+=diffCSC[i];
    }
    seuil = seuil / instance.nb_machines;
    for( i = 0 ; i < instance.nb_machines ; i++ ) {
        if(diffCSC[i]>seuil)
	{
            //DEBUG_PRINT( "machine %u avec diffCSC > seuil\n", i);
	    for( j = 0 ; j < instance.nb_ressources ; j++ ) {
// 	      if(instance.ressources[j].transcient)
                instance.machines[i].SC[j]=instance.machines[i].C[j];
            }
	}
    }  
}

void unmodif_sc(){
    unsigned int i, j;
    /*for( i = 0 ; i < instance.nb_machines ; i++ ) {
        for( j = 0 ; j < instance.nb_ressources ; j++ ) {
            instance.machines[i].SC[j]=instance.machines[i].SCsauv[j];
        }
    }  */
}

/**
 * Lecture d'un fichier contenant l'instance sur laquelle on travaille. 
 * Prend en entree le nom du fichier a lire
 * et un pointeur vers la ou on stocke les donnees.
 * Retourne FMR_OK si tout s'est bien passe, une erreur si il 
 * y a eu un probleme.
 */

int lire_instance( char* filename ){

    FILE* fd;
    unsigned short  j, k, r, m;
    unsigned short i, s;
#ifdef DEBUGSTATS
	unsigned nbTotalDep;
#endif

#ifdef DEBUGIO
    int l;
#endif

    fd = fopen( filename, "r" );
    if( NULL == fd ) {
        perror( "Erreur ouverture en lecture du fichier d\'input " );
        return FMR_IO_ERROR;
    }

    /* --- Ressources --- */

    fscanf( fd, "%hu\n", &instance.nb_ressources );
    DEBUG_IO_PRINT( "%u ressources\n", instance.nb_ressources );

    instance.nb_transcient=0;

    for( r = 0 ; r < instance.nb_ressources ; r++ ) {
        unsigned trans;
        fscanf( fd, "%u %u\n", &trans, & instance.ressources[r].weight_loadCost );
        if( 1 == trans ) {
           instance.ressources[r].transcient = true;
		instance.nb_transcient++;
        } else {
            instance.ressources[r].transcient = false;
        }

        DEBUG_IO_PRINT( "poids de ressource %u: %u ; ", r, instance.ressources[r].weight_loadCost );
        if( instance.ressources[r].transcient ) {
            DEBUG_IO_PRINT( "transcient\n" );
        } else {
            DEBUG_IO_PRINT( "NON transcient\n" );
        }
    }
    
	DEBUG_STATS_PRINT("Nb transcient=%hu\n",instance.nb_transcient);
    /* --- Machines --- */

    fscanf( fd, "%hu\n", & instance.nb_machines );
    DEBUG_IO_PRINT( "%u machines\n", instance.nb_machines );

    instance.machines = (machine_t *)malloc(instance.nb_machines *sizeof(machine_t));

    instance.nb_locations=0;
    instance.nb_neighborhood=0;
    for( i = 0 ; i < instance.nb_machines ; i++ ) {
        fscanf( fd, "%hu %hu ", &(instance.machines[i].neighborhood), &(instance.machines[i].location) );
	if (instance.machines[i].neighborhood >= instance.nb_neighborhood)
		instance.nb_neighborhood = (unsigned short)(instance.machines[i].neighborhood + 1);
	if (instance.machines[i].location >= instance.nb_locations)
		instance.nb_locations = (unsigned short)(instance.machines[i].location+1);
        instance.machines[i].C =  (unsigned *) malloc( instance.nb_ressources * sizeof( unsigned) );  
        instance.machines[i].SC = (unsigned *) malloc( instance.nb_ressources * sizeof( unsigned) );
        //instance.machines[i].SCsauv = (unsigned *) malloc( instance.nb_ressources * sizeof( unsigned) );
        instance.machines[i].MMC = (unsigned*) malloc( instance.nb_machines * sizeof( unsigned) );

        for( j = 0 ; j < instance.nb_ressources ; j++ ) {
            fscanf( fd, "%u ", &instance.machines[i].C[j] );
        }
        for( j = 0 ; j < instance.nb_ressources ; j++ ) {
            fscanf( fd, "%u ", &instance.machines[i].SC[j] );
            //instance.machines[i].SCsauv[j]=instance.machines[i].SC[j];
        }

        for( j = 0 ; j < instance.nb_machines ; j++ ) {
            fscanf( fd, "%u ", &instance.machines[i].MMC[j] );
        }
        
        fscanf( fd, "\n" ); /* juste pour decaler l'offset */
    }
#ifdef DEBUGIO
	DEBUG_IO_PRINT("Matrice des couts des MMC :\n");
	for (i=0; i < instance.nb_machines ; i++ ) {
		for (j=0; j < instance.nb_machines ; j++ ) {
			DEBUG_IO_PRINT( "%10u", instance.machines[i].MMC[j] );
		}
		DEBUG_IO_PRINT("\n");
	}

	DEBUG_IO_PRINT("Contenu des voisinages:\n");
	for (i=0; i< instance.nb_neighborhood;i++) {
		DEBUG_IO_PRINT("n=%3u:",i);
		for(m=0; m< instance.nb_machines;m++) {
			if (instance.machines[m].neighborhood==i)
				DEBUG_IO_PRINT("%3u,",m);
		}
		DEBUG_IO_PRINT("\n");
	}
	
#endif

 //   DEBUG_IO_PRINT("%u locations\n",instance.nb_locations);

   // for (i=0; i< instance.nb_locations; i++)
	    //liste_init(& instance.L[i], instance.nb_machines);

    /* Affectation des ensembles de localisations de machines */
    //for ( m = 0 ; m < instance.nb_machines ; m++ ) {
	    //DEBUG_IO_PRINT("m=%u, location=%u\n",m,instance.machines[m].location);
	    //liste_in(m, &instance.L[instance.machines[m].location]);
    //}
#ifdef DEBUGIO
    //for ( l=0 ; (unsigned) l < instance.nb_locations ; l++) {
       //DEBUG_IO_PRINT("location[%u]=",l);
       //liste_print(&instance.L[l]);
    //}
    //DEBUG_IO_PRINT("\n");
#endif

    /* --- Services --- */

    fscanf( fd, "%hu\n", &instance.nb_services );
    DEBUG_IO_PRINT( "%hu services\n", instance.nb_services );

    instance.services = (service_t*) malloc( instance.nb_services * sizeof( service_t ) );
    //instance.serviceSuccesseur = (unsigned *) malloc( instance.nb_services * sizeof( liste_t) );

    for(s=0;s <instance.nb_services;s++)
	instance.services[s].nb_succ=0;

    for( s = 0 ; s < instance.nb_services ; s++ ) {
        fscanf( fd, "%hu %hu ", &instance.services[s].spreadMin, &instance.services[s].nb_dep );

#ifdef DEBUGIO
        if (instance.services[s].nb_dep>0)
		printf( "Service %4hu depend de %3hu autres services\n", s, instance.services[s].nb_dep );
#endif

        if (instance.services[s].nb_dep>0){
            instance.services[s].dependances = (unsigned short*) malloc( instance.services[s].nb_dep * sizeof(unsigned short) );
        } else {
            instance.services[s].dependances = NULL;
        }
        instance.services[s].nb_procs=0;

        for( j = 0 ; j < instance.services[s].nb_dep ; j++ ) {
            fscanf( fd, "%hu ", &k );
            instance.services[s].dependances[j]=k;
	    instance.services[k].nb_succ++;
            DEBUG_IO_PRINT( "%hu,",instance.services[s].dependances[j]);
        }
        DEBUG_IO_PRINT("\n");
        fscanf( fd, "\n" ); /* juste pour decaler l'offset */
    }
	
    for (s=0; s < instance.nb_services ; s++ ) {
	   instance.services[s].successeurs= (unsigned short *) malloc (instance.services[s].nb_succ * sizeof(unsigned short));
	   instance.services[s].nb_succ=0;
    }
    for (s=0; s < instance.nb_services ; s++ ) {
	for (j = 0 ; j < instance.services[s].nb_dep ; j++ ) {
		k=instance.services[s].dependances[j];
		instance.services[k].successeurs[instance.services[k].nb_succ++]=s;
	}
    }

#ifdef DEBUGIO
   for (s=0; s < instance.nb_services ; s++ ) {
		for(j=0;j< instance.services[s].nb_succ;j++)
		    DEBUG_PRINT("%hu,",instance.services[s].successeurs[j]);
		DEBUG_PRINT("\n");
    }
#endif

    /* --- Processus --- */

    fscanf( fd, "%hu\n", &instance.nb_procs );
    DEBUG_IO_PRINT( "%hu procs\n", instance.nb_procs );

    instance.procs = (processus_t*) malloc( instance.nb_procs * sizeof( processus_t ) );

/*
    for (s=0; s < instance.nb_services ; s++ ) {
    	liste_init(& instance.services[s].ens_procs, instance.nb_procs);
    }
*/
   listeCompacte_alloc(& instance.services_ens_procs, instance.nb_services, instance.nb_procs);
   listeCompacte_init(& instance.services_ens_procs, instance.nb_services, instance.nb_procs);

    for( i = 0 ; i < instance.nb_procs ; i++ ) {
        instance.procs[i].R = (unsigned *) malloc( instance.nb_ressources * sizeof( unsigned) );
        instance.procs[i].id = i;

        fscanf( fd, "%hu  ", &s );
	instance.procs[i].service = s;
	//liste_in(i, & instance.services[s].ens_procs);
	listeCompacte_in(& instance.services_ens_procs, s, i);
	instance.services[s].nb_procs++;
        for( j = 0 ; j < instance.nb_ressources ; j++ ) {
            fscanf( fd, "%u ", &instance.procs[i].R[j] );
        }
        fscanf( fd, "%u\n", &instance.procs[i].PMC );

    }
#ifdef DEBUGSTATS
	i=0;
	nbTotalDep=0;
	for (s=0; s < instance.nb_services; s++) {
		if (instance.services[s].nb_procs==1) {
			if (instance.services[s].nb_dep!=0 || instance.services[s].nb_succ!=0)
				printf("service %hu de card. 1, mais a %hu pred et %u succ\n",s,instance.services[s].nb_dep, instance.services[s].nb_succ);
			i++;
		}
		nbTotalDep += instance.services[s].nb_dep;
	}
	printf ("%hu services de card>1 ; nbTotalDep=%u\n", instance.nb_services - i, nbTotalDep);
#endif

#ifdef DEBUGIO
	DEBUG_IO_PRINT("Liste des ps par service :\n");
	for(s=0; s< instance.nb_services; s++) {
	   DEBUG_IO_PRINT("%4u: (%4u ps) ", s, instance.services[s].nb_procs);
	   liste_print(&instance.services[s].ens_procs);
	   DEBUG_IO_PRINT("\n");
	}
        DEBUG_IO_PRINT("%u voisinages :\n",instance.nb_neighborhood);
	for (i=0; i< instance.nb_neighborhood; i++) {
		DEBUG_IO_PRINT("%2u: ",i);
		for (j=0; j< instance.nb_machines; j++) 
			if (instance.machines[i].neighborhood == j)
				DEBUG_IO_PRINT("%3u,",j);
		DEBUG_IO_PRINT("\n");
	}
#endif

    /* --- Objectifs --- */

    fscanf( fd, "%hu\n", &instance.nb_objectifs );
    DEBUG_IO_PRINT( "%hu objectifs\n", instance.nb_objectifs );
    
    for( i = 0 ; i < instance.nb_objectifs ; i++ ) {
        fscanf( fd, "%hu %hu %u %u\n", &instance.objectifs[i].r1, &instance.objectifs[i].r2, &instance.objectifs[i].target, &instance.objectifs[i].weight_balanceCost );
	DEBUG_IO_PRINT("%hu: r1=%hu, r2=%hu, target=%hu, cout=%u\n",i,instance.objectifs[i].r1, instance.objectifs[i].r2, instance.objectifs[i].target, instance.objectifs[i].weight_balanceCost );
    }

    /* --- Poids --- */

    fscanf( fd, "%u\n", &instance.weight_PMC );
    fscanf( fd, "%u\n", &instance.weight_SMC );
    fscanf( fd, "%u\n", &instance.weight_MMC );

    DEBUG_IO_PRINT( "Poids des couts : process %hu service %hu machine %hu\n", instance.weight_PMC, instance.weight_SMC, instance.weight_MMC );

    /* Ouf ! on est au bout */

    fclose( fd );

    return FMR_OK;
}


/** 
 * Lire la solution initiale dans le fichier.
 * Prend en entree le nom du fichier, le nombre de machines 
 * et un pointeur vers la ou on va stocker les donnees 
 * de la solution initiale. 
 * Retourne FMR_OK si tout s'est bien passe, une erreur si il 
 * y a eu un probleme.
 */

int 
lire_init( char* filename, unsigned short nb_procs, placement_t* solution ){

    FILE* fd;
    unsigned short n,m,r,p,s,l;

    DEBUG_IO_PRINT( "   *** Placement initial : ***\n" );

    fd = fopen( filename, "r" );
    if( NULL == fd ) {
        perror( "Erreur ouverture en lecture du fichier contenant la solution initiale " );
        return FMR_IO_ERROR;
    }

    listeCompacte2D_alloc( & (solution->liste_de_processus_d_un_service_dans_un_voisinage), instance.nb_neighborhood, instance.nb_services, instance.nb_procs);
    listeCompacte2D_init( & (solution->liste_de_processus_d_un_service_dans_un_voisinage), instance.nb_neighborhood, instance.nb_services, instance.nb_procs);

	/* spread est un array2D donc no pb */
    memset(solution->spread[0],0, (size_t)instance.nb_services*instance.nb_locations*sizeof(unsigned short));

    memset(solution->nb_procs_par_mach, 0, instance.nb_machines *sizeof(unsigned short));
    for( p = 0 ; p < nb_procs ; p++ ) {
        fscanf( fd, "%hu", &m);
        solution->assignment[p]=m;
	solution->nb_procs_par_mach[m]++;
	solution->spread[instance.procs[p].service][instance.machines[m].location]++;
	listeCompacte2D_in( &(solution->liste_de_processus_d_un_service_dans_un_voisinage), instance.machines[m].neighborhood, instance.procs[p].service, p);

        DEBUG_IO_PRINT( "Process %4hu sur machine %hu, n=%2hu, s=%4hu, ", p, solution->assignment[p], instance.machines[m].neighborhood, instance.procs[p].service );
        DEBUG_IO_PRINT( "deplt coute %3u, Res=", instance.procs[p].PMC);
#ifdef DEBUGIO
	for (r=0; r  < instance.nb_ressources ; r++) {
		DEBUG_IO_PRINT("%2hu:%7u,",r, instance.procs[p].R[r]);
	}
	DEBUG_IO_PRINT("\n");
#endif 	
    }

#ifdef DEBUGIO
    for (n=0; n<instance.nb_neighborhood; n++)
	for (s=0; s<instance.nb_services;s++) {
		printf("liste[%hu,%hu]=",n,s);
	        listeCompacte2D_print(&(solution->liste_de_processus_d_un_service_dans_un_voisinage),n,s);
	        printf("\n");
	}
#endif 	


    memset(solution->cardServiceMove, 0, instance.nb_services* sizeof(unsigned short));
    solution->maxCardServiceMove=0;

    for( p = 0 ; p < instance.nb_procs ; p++ ) {
        m = solution->assignment[p];
        for( r = 0 ; r < instance.nb_ressources ; r++ ) {
            solution->U[m][r] += instance.procs[p].R[r];
        }
    }

    listeCompacte_init(&solution->liste_procs, instance.nb_machines, instance.nb_procs);
    for (m=0; m< instance.nb_machines; m++) {
	    for (s=0; s<instance.nb_services;s++)
		    solution->service_par_machine[m][s]= NOSERVICE;
    } 
    /* Processus assignes aux machines */
    for( p = 0 ; p < instance.nb_procs ; p++ ) {
        m = solution->assignment[p];
	listeCompacte_in(& solution->liste_procs,m, (unsigned short) p);
        solution->nb_procs_par_mach[m]++;
	solution->service_par_machine[m][instance.procs[p].service]=p;
    }
#ifdef DEBUGIO
    if (capacity_constraint( solution))
	    DEBUG_IO_PRINT("capacity init ok\n");
    else
	    DEBUG_IO_PRINT("pb initial !!!!\n");
    if (conflicts_constraint( solution))
	    DEBUG_IO_PRINT("contraintes de conflits init: ok\n");
    else
	    DEBUG_IO_PRINT("contraintes de conflits init: pb\n");
#endif

    fclose( fd );

    return FMR_OK;
}

/** 
 * Ecriture du resultat dans le fichier resultat.
 * Prend en entree le nom du fichier, le nombre de machines 
 * et un pointeur vers la ou on va stocker les donnees 
 * de la solution optimisee. 
 * Retourne FMR_OK si tout s'est bien passe, une erreur si il 
 * y a eu un probleme.
 */

int ecrire_resultat( char* filename , placement_t* solution){

    FILE* fd;
    unsigned int i;

    fd = fopen( filename, "w" );
    if( NULL == fd ) {
        perror( "Erreur ouverture en ecriture du fichier resultat " );
        return FMR_IO_ERROR;
    }

    for( i = 0 ; i < instance.nb_procs ; i++ ) {
        fprintf( fd, "%u ", solution->assignment[i] );
    }
    fprintf( fd, "\n" );

    fclose( fd );
    return FMR_OK;
}

int ecrire_struct_resultat (char * filename, placement_t* solution, unsigned int maxcomp, unsigned int maxcomp2, int64 init_cout, int64 final_cout){
    FILE* fd;
    unsigned int r,m;
    liste_t *l;

    fd = fopen(filename,"w");
    if ( NULL == fd) {
	perror( "Erreur ouverture en ecriture du fichier structure du resultat");
	return FMR_IO_ERROR;
    }
   
    fprintf(fd, "init= %llu, final=%llu\n",init_cout, final_cout); 
    fprintf(fd, "Nb proc= %u, Nb mach=%u,",instance.nb_procs, instance.nb_machines);
    fprintf(fd, "Nb services= %u, Nb ressources=%u,", instance.nb_services, instance.nb_ressources);
    fprintf(fd, "Nb locations= %u, Nb neighborhood=%u", instance.nb_locations, instance.nb_neighborhood);
    fprintf(fd, "Nb objectifs= %u\n", instance.nb_objectifs);
    fprintf(fd, "comp. connexe max=%u, 2e max=%u\n",maxcomp, maxcomp2);
    for (m=0; m <instance.nb_machines; m++) {
        fprintf (fd, "%u:",m);
	listeCompacte_fprint(fd, & solution->liste_procs, m);
	for (r=0; r< instance.nb_ressources; r++) {
		fprintf(fd, "R[%u]=%u/%u(%.2f %%)/%u(%.2f %%),",r, solution->U[m][r], 
			instance.machines[m].SC[r], 100.0*instance.machines[m].SC[r]/ instance.machines[m].C[r],
 			instance.machines[m].C[r], 100.0*solution->U[m][r]/ instance.machines[m].C[r]);
	}
	fprintf(fd,"\n");
    }
    fclose(fd);
    return FMR_OK;
}

/* Affichage d'un assignement */
void print_assign( unsigned short nb_procs, placement_t* placement ) {
    unsigned short i;
    for( i = 0 ; i < nb_procs ; i++ ) {
        printf( "%hu:%hu   ", i, placement->assignment[i] );
    }
    printf( "\n" );
}
