/**********************************************************************
 * This file is part of FMR (Fast Machine Reassignment)               
 *                                                                   
 * ROADEF/EURO Challenge 2012: Machine Reassignment                 
 * http://challenge.roadef.org/2012/en/				   
 * Team S26
 *     
 * Copyright (C) 2011-12 Laboratoire d'Informatique de Paris Nord 
 * UMR CNRS 7030 
 * Institut Galilée - Université Paris-Nord
 * 99, avenue Jean-Baptiste Clément 
 * 93430 Villetaneuse  
 *        
 * FMR is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *  
 * FMR is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with FMR; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **********************************************************************/
#include "fmr_threads.h"
#include "fmr_mem.h"


struct atomic_resultat {
    placement_t resultats[2] ;
    int idx_resultat ;
    pthread_mutex_t mutex;
};

typedef struct atomic_resultat threaded_placement_t ;

static threaded_placement_t atomic_meilleur_placement ;


void
atomic_init_meilleur_placement()
{
    atomic_meilleur_placement.idx_resultat = 0;
    atomic_meilleur_placement.resultats[0] = alloc_assignement (instance.nb_procs);
    atomic_meilleur_placement.resultats[1] = alloc_assignement (instance.nb_procs);
    pthread_mutex_init(&atomic_meilleur_placement.mutex , NULL );
}


placement_t *
atomic_get_meilleur_placement()
{
    return atomic_meilleur_placement.resultats + atomic_meilleur_placement.idx_resultat ;
}


void
atomic_set_meilleur_placement(placement_t * proposition, char algo)
{
    int idx_dispo = 1 - atomic_meilleur_placement.idx_resultat ;
    placement_t * dispo = atomic_meilleur_placement.resultats + idx_dispo ;
    
#ifdef TRACE_ALGOS
	if ( 'a'>algo || algo>'z' || proposition->nb_algos==0 || proposition->liste_algos[proposition->nb_algos-1]!=algo ) {
	    proposition->liste_algos = (char*)realloc(proposition->liste_algos, sizeof(char)*((long unsigned)proposition->nb_algos+2));
	    proposition->liste_algos[proposition->nb_algos]=algo;
	    proposition->liste_algos[++proposition->nb_algos]=0;
	}
 	//printf("coût : %15ld   algorithmes utilisés : %s\n", proposition->cout, proposition->liste_algos);
#endif

	// section critique
    pthread_mutex_lock (&atomic_meilleur_placement.mutex);
    if (proposition->cout < atomic_meilleur_placement.resultats[atomic_meilleur_placement.idx_resultat].cout) {
	//printf("Amélioration par %lu : %lu\n", pthread_self(), proposition->cout );
        // copie du nouveau meilleur résultat (non atomique)
        copier_assignement( proposition, dispo );
        // changement (atomique) de l'index du meilleur résultat
        atomic_meilleur_placement.idx_resultat = idx_dispo ;
#ifdef TRACE_ALGOS
 	printf("coût : %15lld   algorithmes utilisés : %s\n", proposition->cout, proposition->liste_algos);
#endif
    }
    pthread_mutex_unlock(&atomic_meilleur_placement.mutex);
}


void atomic_if_meilleur_placement_copyto(placement_t* proposition)
{
    // section critique
    pthread_mutex_lock (&atomic_meilleur_placement.mutex);
    if (proposition->cout > atomic_meilleur_placement.resultats[atomic_meilleur_placement.idx_resultat].cout) {
        // copie du nouveau meilleur résultat (non atomique)
        copier_assignement( atomic_meilleur_placement.resultats + atomic_meilleur_placement.idx_resultat, proposition);
    }
    pthread_mutex_unlock(&atomic_meilleur_placement.mutex);
}

int64_t atomic_get_meilleur_cout()
{
    pthread_mutex_lock (&atomic_meilleur_placement.mutex);
    int64_t c = atomic_meilleur_placement.resultats[atomic_meilleur_placement.idx_resultat].cout;
    pthread_mutex_unlock(&atomic_meilleur_placement.mutex);
    return c;
}

void
atomic_free_placement()
{
    free_placement(atomic_meilleur_placement.resultats);
    free_placement(atomic_meilleur_placement.resultats+1);
    pthread_mutex_destroy(&atomic_meilleur_placement.mutex);
}

void atomic_lock_placement()
{
    pthread_mutex_lock (&atomic_meilleur_placement.mutex);
}

void atomic_unlock_placement()
{
    pthread_mutex_unlock(&atomic_meilleur_placement.mutex);
}
