/**********************************************************************
 * This file is part of FMR (Fast Machine Reassignment)               
 *                                                                   
 * ROADEF/EURO Challenge 2012: Machine Reassignment                 
 * http://challenge.roadef.org/2012/en/				   
 * Team S26
 *     
 * Copyright (C) 2011-12 Laboratoire d'Informatique de Paris Nord 
 * UMR CNRS 7030 
 * Institut Galilée - Université Paris-Nord
 * 99, avenue Jean-Baptiste Clément 
 * 93430 Villetaneuse  
 *        
 * FMR is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *  
 * FMR is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with FMR; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **********************************************************************/
#ifndef fmr_tools
# define fmr_tools 1

#include "fmr.h"

/* variables globales */
unsigned long long tempsZero;
struct timeval tv_tempsInit;


void init_rand( unsigned int seed, struct drand48_data * rand_buf);

static inline double frand_a_b(struct drand48_data * rand_buf, double a, double b) {
    double rval;
    return ( drand48_r(rand_buf, &rval),rval ) * (b-a) + a;
}
static inline unsigned irand_a_b(struct drand48_data * rand_buf, unsigned a,unsigned b) {
    long rval;
    return  (unsigned) (a + (lrand48_r(rand_buf, &rval),(unsigned) rval)%((unsigned)(b-a)) );
}



processus_t* delister( struct process_list*, unsigned int );
struct process_list* enlister( struct process_list*, processus_t* );

# define LISTE_VIDE  -1u
# define LISTE_VIDE_US  	((unsigned short)-1u)

unsigned short liste_out(unsigned short, liste_t*);
unsigned short liste_alea(liste_t *, struct drand48_data * rand_buf);
void liste_in(unsigned short, liste_t*);
void liste_init(liste_t*, unsigned short);
void liste_dup (liste_t*, const liste_t*, unsigned short);
void liste_cpy (liste_t*, const liste_t*, unsigned short);
void liste_print(liste_t*);
void liste_fprint(FILE *, liste_t*);
void liste_free(liste_t*);

unsigned short liste_next(unsigned short, liste_t *);
unsigned short liste_head(liste_t *);
void liste_remplie(liste_t *l);
void liste_efface(liste_t *l);
bool liste_appartient(unsigned short, liste_t *);
bool liste_est_vide(liste_t *);
bool liste_est_singleton(liste_t *);

/*** traitement de listeCompacte ***/
void listeCompacte_alloc(listeCompacte_t *, unsigned, unsigned );
void listeCompacte_init(listeCompacte_t *l, unsigned dimIndex, unsigned dimChainage);
void listeCompacte_in(listeCompacte_t *, unsigned ,  unsigned short);
void listeCompacte_out(listeCompacte_t *, unsigned , unsigned short);
void listeCompacte_fprint(FILE *, listeCompacte_t *, unsigned );
void listeCompacte_free(listeCompacte_t *);
void listeCompacte_cpy (listeCompacte_t *dest, const listeCompacte_t const *src, unsigned dimIndex, unsigned dimChainage);
bool listeCompacte_est_vide(listeCompacte_t *, unsigned );
bool listeCompacte_est_singleton(listeCompacte_t *, unsigned );

unsigned short listeCompacte_long(const listeCompacte_t *l, unsigned c) ;
unsigned short listeCompacte_alea(const listeCompacte_t *l, unsigned c, struct drand48_data * rand_buf) ;

unsigned short listeCompacte_tete(const listeCompacte_t *, unsigned c) ;
unsigned short listeCompacte_next(const listeCompacte_t *, unsigned c, unsigned short i) ;

/*** traitement de listeCompacte2D ***/
void listeCompacte2D_alloc(listeCompacte2D_t *, unsigned int, unsigned int, unsigned int);
void listeCompacte2D_init(listeCompacte2D_t *, unsigned int, unsigned int, unsigned int);
void listeCompacte2D_in(listeCompacte2D_t *, unsigned int, unsigned int, unsigned int);
void listeCompacte2D_out(listeCompacte2D_t *, unsigned int, unsigned int, unsigned int);
void listeCompacte2D_print(listeCompacte2D_t *, unsigned int, unsigned int);
void listeCompacte2D_free(listeCompacte2D_t *);
void listeCompacte2D_cpy(listeCompacte2D_t *, const listeCompacte2D_t const *,  unsigned int, unsigned int, unsigned int);
bool listeCompacte2D_est_vide(listeCompacte2D_t *, unsigned int, unsigned int);
bool listeCompacte2D_est_singleton(listeCompacte2D_t *, unsigned int, unsigned int);
void listeCompacte2D_fprint(FILE *f, listeCompacte2D_t *l, unsigned int c1, unsigned int c2) ;


bool verifCardServices (const placement_t const *);

/*** traitement des array2D */
void **array2Dmalloc(size_t , unsigned , unsigned int ) ;
void **array2Dcalloc(size_t , unsigned , unsigned int ) ;
void array2Dcpy (void *, const void *, size_t , unsigned int , unsigned int ) ;
void array2Dmemset(void **ptr, int c, size_t s, unsigned nb1D, unsigned int nb2D) ;

/*** gestion temps restant et mémoire restante */
void tempsInit();
void tempsInit2();
float tempsRestant(unsigned time_limit);
double tempsRestant2() ;
float tempsEcoule();
double tempsEcoule2();
unsigned long memoireRestante();
void printMemoireUtilisee() ;

float cpufreq;
extern unsigned int time_limit;
#endif
